

#define PID 0

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

volatile int pwm,dd;
volatile float Vsp,Vin,Vout,xVin,xVout,Iout,Isp,xIsp,xVsp,ifactor;
volatile float xIout, hs_temp;
volatile int oIout;
volatile float Input, Output, Setpoint,fpwm, errSum, lastErr,error,dErr, kp, ki, kd;
volatile float iInput, iOutput, iSetpoint, ierrSum, ilastErr,ierror,idErr, ikp, iki, ikd;
volatile int f50c,f50;

volatile uint8_t blcd,go_lcd = 0,bpos=0,baud_waits = 1;
volatile int baudc = 0,i_timeout = 0,dc=0;
volatile char buf[85],*bp,bcount;
volatile float amp_hr = 0.0;
volatile long old_mil;

#define PWM 799       // PWM freq is 20kHz
#define FPWM 799.0

void setup()
  {  
  //Serial.begin(9600);
    
  pwm = 1;
  ifactor = 1.0;
  pinMode(9,OUTPUT);
  digitalWrite(9,0);
  
  bcount = 0;             
  bp = &buf[0];
  sprintf(buf,"booting..");
  
  pinMode(4,OUTPUT);
  pinMode(5,OUTPUT);
  TCCR1A = _BV(COM1A1) | _BV(COM1B1) | _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  OCR1A = 1;
  ICR1 = PWM;
  TIMSK1 |= (1 << TOIE1);   // enable timer overflow interrupt
  delay(500);
  dd = 0;
  f50c = 0;
  f50 = 0;
  
  // cals 
  xVin = 0.1199;
  xVout = 0.11156;
  xIout = 0.0631;
  oIout = 104;
  xIsp = 0.020;
  xVsp = 0.088;
  
  //pid
  kp = 0.1;     // set up PID for  output voltage control   0.1 0.0 0.01
  ki = 0.0;
  kd = 0.02;
  errSum = 0.0;        // initialise PID
  lastErr = 0.0;

  ikp = 0.1;     // set up PID for output current control   0.1 0.0 0.01
  iki = 0.0;
  ikd = 0.01;
  ierrSum = 0.0;        // initialise PID
  ilastErr = 0.0;
  
  fpwm = 5.0;
  
  bcount = 0;             
  bp = &buf[0]; 
  old_mil = millis();
  
 }

ISR(TIMER1_OVF_vect)        
{
OCR1A = pwm;    // pin 9

// send one char out pin A4. First, get a character and put it in "blcd" and init state machine.
if (go_lcd == 0 && baudc > 50)    // not send as soon as possible, let the LCD digest the data with a delay here
  {
  // get next char
  blcd = buf[bcount];
  bcount++;
  if (bcount > 80)
    {
    bcount = 0;
    blcd =  '$';     // this is translated by the serial LCD as "put cursor at top left position"
    }   
  go_lcd = 1;
  bpos = 0;
  baudc = 2;
  }
  
if (go_lcd == 1 && baudc > baud_waits)   // a little 8 bit serial output state machine
  {
    if(bpos == 9)
    {
    // send stop bit, pull high
    sbi(PORTD,4);
    go_lcd = 0;
    baudc = 0;
    bpos = 0;
    }
  if (bpos > 0  )
    {
      if (blcd & 1)
        sbi(PORTD,4);
      else
        cbi(PORTD,4);
      blcd = blcd >> 1;
      bpos++;  
      baudc = 0;
    }
  if(bpos == 0 && go_lcd == 1)
    {
    // send start bit, pull low
    cbi(PORTD,4);
    bpos++;  
    baudc = 0;
    }
  }
  baudc++;

f50c++;
if(f50c > 49)
  {
    f50c=0;
    f50=1;
  }
}   


void do_pid()
{   
   float timeChange = 0.01;    // always called at 100Hz, so dT will be 10msec..
   error = Setpoint - Input;
   errSum += (error * timeChange);
   dErr = (error - lastErr) / timeChange;
   Output = kp * error + ki * errSum + kd * dErr;
   lastErr = error;
}

void do_ipid()
{   
   float timeChange = 0.01;    // always called at 100Hz, so dT will be 10msec..
   ierror = iSetpoint - iInput;
   ierrSum += (ierror * timeChange);
   idErr = (ierror - ilastErr) / timeChange;
   iOutput = ikp * ierror + iki * ierrSum + ikd * idErr;
   ilastErr = ierror;
}

float get_temp(int Vo,int hi_low)
{
 //float a = 1.027957148e-3, b = 2.393569459e-4, c = 1.555071887e-7;
 float a = 1.009249522e-03, b = 2.378405444e-04, c = 2.019202697e-07;
 float T,logRt,Tc,v,r,i;                                 // assumes NTC on high side of voltage divider
 v = 5.0 * (float)Vo /1024.0;    // v in volts
 if (hi_low == 1)     // NTC is on high side of divider, with 10K on low side
  {
    i = v/1e4;
    r = (5.0 - v)/i;
  }
 else                 // NTC is on low side of divider with 10K on high side
  {
    i = (5.0 - v)/1e4;
    r = v/i;
  }
 logRt = log(r); 
 T = (1.0 / (a + b*logRt + c*logRt*logRt*logRt)); // We get the temperature value in Kelvin from this Stein-Hart equation
 Tc = T - 273.15;                                 // Convert Kelvin to Celsius
 return Tc;
}

   
void loop()
   {
   int nupwm;
   char fbuf[10],hbuf[15],gbuf[10];
   float watts,dtime;
   
   Isp = xIsp * (float)analogRead(7); 
   Vsp = xVsp * (float)analogRead(6); 
   Vin = xVin * (float)analogRead(1);
   Vout = xVout * (float)analogRead(0);
   Iout = xIout * (float)(analogRead(3) - oIout);  
   hs_temp = get_temp(analogRead(2),0);
   if (hs_temp > 60.0)    sbi(PORTD,5);
   if (hs_temp < 50.0)    cbi(PORTD,5);  

   if (f50 == 1)
    {
    f50 = 0; 
     
if (PID)            // top line of code will define this = 0 or 1
  {
    if (Iout > Isp)
      {
        ifactor -= 0.01;
        if (ifactor < 0.001) ifactor = 0.001;
      }
    else
      {
        ifactor += 0.01;
        if (ifactor > 1.0) ifactor = 1.0;
      }
      
            Input = Vout;
            Setpoint = Vsp * ifactor;
            do_pid();
            if (Output > 30.0) Output = 30.0;
            if (Output < -30.0) Output = -30.0;
           fpwm += Output;
   }
  else
    {
      if (Vout < Vsp && Iout < Isp) fpwm += 0.50;
      if (Vout > Vsp || Iout > Isp) fpwm -= 0.50;
    }
      
      
      if (fpwm > FPWM) fpwm = FPWM;
      if (fpwm < 1.0) fpwm = 1.0;
      nupwm = int(fpwm);
      if (nupwm > PWM) nupwm = PWM;
      if (nupwm < 1) nupwm = 1;
      if (Vin < Vsp ) {nupwm = 1;fpwm = 1.0;}
      pwm = nupwm;  
    }
   
   if (dd++ > 100)
    {
      dd=0; 
      dc++;
      if (dc > 3) dc = 0;
      if (dc == 0)
        {
        dtostrf(Vin,5,1,fbuf); 
        dtostrf(hs_temp,4,1,hbuf);
        sprintf(bp,"In  %5sV  %4sC ",fbuf,hbuf);
        }
      if (dc == 1)
        {
        dtostrf(Vsp,5,1,fbuf);  
        dtostrf(Isp,4,1,hbuf);
        sprintf(bp+40,"Set %5sV %4sA   ",fbuf,hbuf);
        }
      if (dc == 2)
        {  
        dtostrf(Vout,5,1,fbuf);  
        dtostrf(Iout,4,1,hbuf);
        sprintf(bp+20,"Out %5sV %4sA   ",fbuf,hbuf);
        }
      if (dc == 3)
        {  
        dtostrf(Vout*Iout,6,1,fbuf);  
        dtostrf(amp_hr,4,2,hbuf);
        dtime = (float)(millis() - old_mil);
        old_mil = millis();
        amp_hr += Iout * dtime * 0.001 / 3600.0;
        sprintf(bp+60,"%6sW  %4sA.hr  ",fbuf,hbuf);
        }
    }  
   }

 
